from concurrent import futures
import argparse
import itertools
import functools
import grpc
from . import kama_pb2
from . import kama_pb2_grpc
from google.protobuf.empty_pb2 import Empty
import time
import os

import kama.database
import kama.context
import kama.log


log = kama.log.get_logger('kama.server')


def entity_to_pb(context, entity, detail=True):
    pb = kama_pb2.Entity()
    pb.uuid = entity.uuid
    pb.name = entity.name
    pb.kind = entity.kind

    if detail:
        for attribute in entity.attributes(context=context):
            attribute_pb = pb.attributes.add()
            attribute_pb.uuid = attribute.uuid
            attribute_pb.entity.uuid = attribute.entity_uuid
            attribute_pb.entity.kind = entity.kind
            attribute_pb.entity.name = entity.name
            attribute_pb.key = attribute.key
            attribute_pb.value = attribute.value

        for link in entity.links_from(context=context):
            link_pb = pb.links_from.add()
            link_pb.uuid = link.uuid
            link_pb.from_entity.uuid = link.from_entity.uuid
            link_pb.from_entity.kind = link.from_entity.kind
            link_pb.from_entity.name = link.from_entity.name
            link_pb.to_entity.uuid = link.to_entity.uuid
            link_pb.to_entity.kind = link.to_entity.kind
            link_pb.to_entity.name = link.to_entity.name

        for link in entity.links_to(context=context):
            link_pb = pb.links_to.add()
            link_pb.uuid = link.uuid
            link_pb.from_entity.uuid = link.from_entity.uuid
            link_pb.from_entity.kind = link.from_entity.kind
            link_pb.from_entity.name = link.from_entity.name
            link_pb.to_entity.uuid = link.to_entity.uuid
            link_pb.to_entity.kind = link.to_entity.kind
            link_pb.to_entity.name = link.to_entity.name

        for permission in entity.permissions(context=context):
            permission_pb = pb.permissions.add()
            permission_pb.uuid = permission.uuid
            permission_pb.role.uuid = permission.role.uuid
            permission_pb.role.kind = permission.role.kind
            permission_pb.role.name = permission.role.name
            permission_pb.entity.uuid = permission.entity.uuid
            permission_pb.entity.kind = permission.entity.kind
            permission_pb.entity.name = permission.entity.name
            permission_pb.name = permission.name

    return pb


class DatabaseServicer(kama_pb2_grpc.KamaDatabaseServicer):
    def ListEntities(self, entity, context):
        request_context = kama.context.get_request_context(context)

        # Emulate the behavior of GetEntity. There can be only one.
        if entity.uuid:
            yield kama.database.Entity(entity.uuid)
            return
        if entity.name:
            yield kama.database.Entity.get_by_name(entity.name)
            return

        result = []
        filters = []

        if entity.kind:
            result = kama.database.Entity.get_by_kind(entity.kind)
        else:
            # This is inefficient. We're gonna get *all* the entities and
            # filter them
            result = kama.database.Entity.get_all()

        for attribute in entity.attributes:
            # Why are you searching by attribute uuid? You already know the entity at that point.
            assert not attribute.uuid
            if attribute.key and attribute.value:
                filters.append(lambda x: [y for y in x.attributes(key=attribute.key, context=request_context) if y.value == attribute.value])
            elif attribute.key:
                filters.append(lambda x: x.attributes(key=attribute.key, context=request_context))

        for link in entity.links_from:
            filters.append(lambda x: [y for y in x.links_from(context=request_context) if y.entity_to(context=request_context).uuid == link.to_entity.uuid])

        for link in entity.links_to:
            filters.append(lambda x: [y for y in x.links_to(context=request_context) if y.entity_from(context=request_context).uuid == link.from_entity.uuid])

        for permission in entity.permissions:
            if permission.uuid:
                filters.append(lambda x: [y for y in x.permissions(context=request_context) if y.uuid == permission.uuid])
            if permission.role:
                filters.append(lambda x: [y for y in x.permissions(context=request_context) if y.role.uuid == permission.role.uuid])
            if permission.entity:
                filters.append(lambda x: [y for y in x.permissions(context=request_context) if y.entity.uuid == permission.entity.uuid])
            if permission.name:
                filters.append(lambda x: [y for y in x.permissions(context=request_context) if y.name == permission.name])

        combo_filter = lambda x: all([bool(y(x)) for y in filters])
        for entity in result:
            try:
                if combo_filter(entity):
                    yield entity_to_pb(request_context, entity, detail=False)
            except kama.database.PermissionDeniedException:
                pass

    def GetEntity(self, entity, context):
        request_context = kama.context.get_request_context(context)

        if entity.uuid:
            result = kama.database.Entity(entity.uuid)
        elif entity.kind and entity.name:
            result = kama.database.Entity.get_by_name(entity.kind, entity.name)

        if result:
            return entity_to_pb(request_context, result, detail=True)
        else:
            context.set_code(grpc.StatusCode.NOT_FOUND)
            return entity

    def CreateEntity(self, request, context):
        request_context = kama.context.get_request_context(context)
        owner = kama.database.Entity(request.owner_role.uuid)

        if not owner in [x.from_entity for x in request_context.user.links_to(context=request_context)]:
            # It's a bad idea to create entities owned by roles that you're not
            # a member of, because then you cannot get or modify the new
            # entity. We disallow this to prevent foot shooting.
            context.set_code(grpc.StatusCode.FAILED_PRECONDITION)
            return kama_pb2.Entity()

        entity = kama.database.Entity.create(request.entity.kind, request.entity.name, owner)
        return entity_to_pb(request_context, entity, detail=True)

    def DeleteEntity(self, entity, context):
        request_context = kama.context.get_request_context(context)
        entity = kama.database.Entity(entity.uuid)
        entity.delete(context=request_context)
        return Empty()

    def UpdateEntity(self, entity, context):
        request_context = kama.context.get_request_context(context)
        db_entity = kama.database.Entity(entity.uuid)
        db_entity.set_name(entity.name, context=request_context)
        return entity_to_pb(request_context, db_entity, detail=True)

    def AddAttribute(self, pb_attribute, context):
        request_context = kama.context.get_request_context(context)
        entity = kama.database.Entity(pb_attribute.entity.uuid)
        attribute = entity.add_attribute(key=pb_attribute.key, value=pb_attribute.value, context=request_context)
        pb_attribute.uuid = attribute.uuid
        return pb_attribute

    def DeleteAttributes(self, pb_attribute, context):
        request_context = kama.context.get_request_context(context)
        entity = kama.database.Entity(pb_attribute.entity.uuid)
        entity.delete_attributes(key=pb_attribute.key, context=request_context)
        return Empty()

    def AddLink(self, pb_link, context):
        request_context = kama.context.get_request_context(context)
        from_entity = kama.database.Entity(pb_link.from_entity.uuid)
        to_entity = kama.database.Entity(pb_link.to_entity.uuid)
        link = from_entity.add_link(to_entity, context=request_context)
        pb_link.uuid = link.uuid
        return pb_link

    def DeleteLink(self, pb_link, context):
        request_context = kama.context.get_request_context(context)
        from_entity = kama.database.Entity(pb_link.from_entity.uuid)
        to_entity = kama.database.Entity(pb_link.to_entity.uuid)
        from_entity.delete_link(to_entity, context=request_context)
        return Empty()

    def AddPermission(self, pb_permission, context):
        request_context = kama.context.get_request_context(context)
        role = kama.database.Entity(pb_permission.role.uuid)
        entity = kama.database.Entity(pb_permission.entity.uuid)
        permission = entity.add_permission(role, pb_permission.name, context=request_context)
        pb_permission.uuid = permission.uuid
        return pb_permission

    def DeletePermission(self, pb_permission, context):
        request_context = kama.context.get_request_context(context)
        role = kama.database.Entity(pb_permission.role.uuid)
        entity = kama.database.Entity(pb_permission.entity.uuid)
        entity.delete_permission(role, pb_permission.name, context=request_context)
        return Empty()


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--init-schema', action='store_true')
    parser.add_argument('--server-cert', default=os.environ.get('KAMA_SERVER_CERT', 'server.cert'))
    parser.add_argument('--server-key', default=os.environ.get('KAMA_SERVER_KEY', 'server.key'))
    parser.add_argument('--ca-cert', default=os.environ.get('KAMA_CA_CERT', 'ca-cert.pem'))
    parser.set_defaults(func=main)
    args = parser.parse_args()

    if args.init_schema:
        kama.database.schema_init()
        return

    server_cert = open(args.server_cert, 'r').read()
    server_key = open(args.server_key, 'r').read()
    root_cert = open(args.ca_cert, 'r').read()
    creds = grpc.ssl_server_credentials([(server_key, server_cert)], root_cert, require_client_auth=True)

    servicer = DatabaseServicer()

    server = grpc.server(futures.ThreadPoolExecutor(max_workers=100))
    kama_pb2_grpc.add_KamaDatabaseServicer_to_server(servicer, server)
    server.add_secure_port('[::]:8443', creds)
    server.start()

    try:
        while True:
            time.sleep(60)
    except KeyboardInterrupt:
        server.stop(0)



import logging
import os


# Completely reset default logging config
ROOT = logging.getLogger()
map(ROOT.removeHandler, list(ROOT.handlers))
map(ROOT.removeFilter, list(ROOT.handlers))


LOG_FORMAT = logging.Formatter(
        fmt='%(asctime)s %(levelname)s %(threadName)s %(name)s %(filename)s:%(lineno)d %(message)s',
        datefmt='%Y-%m-%d %H:%M:%SZ')


STDERR = logging.StreamHandler()
STDERR.setFormatter(LOG_FORMAT)
STDERR.setLevel(logging.DEBUG)
ROOT.addHandler(STDERR)

if bool(os.environ.get('KAMA_DEBUG', False)):
    ROOT.setLevel(logging.DEBUG)
else:
    ROOT.setLevel(logging.INFO)


def get_logger(name):
    log = logging.getLogger(name)
    return log



# -*- coding: utf-8 -*-

DEFAULT_ACLS = {
    'readonly': [
        'entity.read_attribute',
        'entity.read_link',
        'entity.read_permission',
    ],
    'readwrite': [
        'entity.read_attribute',
        'entity.add_attribute',
        'entity.delete_attribute',
        'entity.read_link',
        'entity.add_link',
        'entity.delete_link',
        'entity.read_permission',
        'entity.set_name',
    ],
    'all': [
        'entity.delete',
        'entity.read_attribute',
        'entity.add_attribute',
        'entity.delete_attribute',
        'entity.read_link',
        'entity.add_link',
        'entity.delete_link',
        'entity.read_permission',
        'entity.add_permission',
        'entity.delete_permission',
        'entity.set_name',
    ],
}



#!/usr/bin/env python
from google.protobuf.empty_pb2 import Empty
import grpc

from kama.acl import DEFAULT_ACLS
from kama import kama_pb2
from kama import kama_pb2_grpc
import kama.env
import kama.log

import argparse


log = kama.log.get_logger('kama.client')


class KamaDatabaseClient(object):
    def __init__(self, args):
        if args.ca_cert is not None:
            root_pem = open(args.ca_cert, 'r').read()
            client_private_pem = open(args.client_key, 'r').read()
            client_public_pem = open(args.client_cert, 'r').read()

            creds = grpc.ssl_channel_credentials(root_pem, client_private_pem, client_public_pem)
            channel = grpc.secure_channel(args.server, creds)
        else:
            log.warning('Connecting database client over insecure channel')
            channel = grpc.insecure_channel(args.server)
        
        self.stub = kama_pb2_grpc.KamaDatabaseStub(channel)

    def list_entities(self, kind=None):
        query = kama_pb2.Entity()
        if kind:
            query.kind = kind
        return self.stub.ListEntities(query)

    def get_entity(self, kind, name):
        entity = kama_pb2.Entity()
        entity.kind = kind
        entity.name = name
        return self.stub.GetEntity(entity)
    
    def get_entity_by_uuid(self, uuid):
        entity = kama_pb2.Entity()
        entity.uuid = uuid
        return self.stub.GetEntity(entity)

    def create_entity(self, kind, name, owner_role_name):
        request = kama_pb2.CreateEntityRequest()
        request.entity.kind = kind
        request.entity.name = name

        owner_role = self.get_entity('role', owner_role_name)
        request.owner_role.uuid = owner_role.uuid
        return self.stub.CreateEntity(request)
    
    def delete_entity(self, uuid):
        entity = kama_pb2.Entity()
        entity.uuid = uuid
        return self.stub.DeleteEntity(entity)

    def update_entity(self, uuid, name):
        entity = kama_pb2.Entity()
        entity.uuid = uuid
        entity.name = name
        new_entity = self.stub.UpdateEntity(entity)
        return new_entity

    def add_attribute(self, entity_uuid, key, value):
        attribute = kama_pb2.Attribute()
        attribute.entity.uuid = entity_uuid
        attribute.key = key
        attribute.value = value
        return self.stub.AddAttribute(attribute)

    def delete_attributes(self, entity_uuid, key):
        attribute = kama_pb2.Attribute()
        attribute.entity.uuid = entity_uuid
        attribute.key = key
        return self.stub.DeleteAttributes(attribute)

    def add_link(self, from_uuid, to_uuid):
        link = kama_pb2.Link()
        link.from_entity.uuid = from_uuid
        link.to_entity.uuid = to_uuid
        return self.stub.AddLink(link)

    def delete_link(self, from_uuid, to_uuid):
        link = kama_pb2.Link()
        link.from_entity.uuid = from_uuid
        link.to_entity.uuid = to_uuid
        return self.stub.DeleteLink(link)

    def add_permission(self, entity_uuid, role_uuid, name):
        permission = kama_pb2.Permission()
        permission.entity.uuid = entity_uuid
        permission.role.uuid = role_uuid
        permission.name = name
        return self.stub.AddPermission(permission)

    def delete_permission(self, entity_uuid, role_uuid, name):
        permission = kama_pb2.Permission()
        permission.entity.uuid = entity_uuid
        permission.role.uuid = role_uuid
        permission.name = name
        return self.stub.DeletePermission(permission)

    def pb_to_human(self, entity):
        out = '''name:    {entity.name}
kind:    {entity.kind}
uuid:    {entity.uuid}
attributes:\n'''.format(entity=entity)

        for attribute in entity.attributes:
            out += '    {attribute.key} = {attribute.value}\n'.format(attribute=attribute)

        out += 'links:\n'
        for link in entity.links_from:
            out += '    -> %s %s\n' % (link.to_entity.kind, link.to_entity.name)

        for link in entity.links_to:
            out += '    <- %s %s\n' % (link.from_entity.kind, link.from_entity.name)

        out += 'permissions:\n'
        for permission in entity.permissions:
            out += '    %s %s\n' % (permission.role.name, permission.name)

        return out


def list_entities(args):
    client = KamaDatabaseClient(args)
    for entity in client.list_entities(kind=args.kind):
        print('%s %s' % (entity.kind, entity.name))

def get_entity(args):
    client = KamaDatabaseClient(args)
    print(client.pb_to_human(client.get_entity(args.kind, args.name)))

def create_entity(args):
    client = KamaDatabaseClient(args)
    entity = client.create_entity(args.kind, args.name, args.owner_role)
    print(client.pb_to_human(entity))

def delete_entity(args):
    client = KamaDatabaseClient(args)
    entity = client.get_entity(args.kind, args.name)
    client.delete_entity(entity.uuid)

def rename_entity(args):
    client = KamaDatabaseClient(args)
    entity = client.get_entity(args.kind, args.old_name)
    entity = client.update_entity(entity.uuid, args.new_name)
    print(client.pb_to_human(entity))


def attribute_add(args):
    client = KamaDatabaseClient(args)
    entity = client.get_entity(args.entity_kind, args.entity_name)
    client.add_attribute(entity.uuid, args.key, args.value)

def attribute_delete(args):
    client = KamaDatabaseClient(args)
    entity = client.get_entity(args.entity_kind, args.entity_name)
    client.delete_attributes(entity.uuid, args.key)

def attribute_list(args):
    client = KamaDatabaseClient(args)
    entity = client.get_entity(args.entity_kind, args.entity_name)

    if args.key is not None:
        attributes = [x for x in entity.attributes if x.key == args.key]
    else:
        attributes = entity.attributes

    for attribute in attributes:
        print('%s=%s' % (attribute.key, attribute.value.encode('utf-8')))


def link_add(args):
    client = KamaDatabaseClient(args)
    from_entity = client.get_entity(args.entity_from_kind, args.entity_from_name)
    to_entity = client.get_entity(args.entity_to_kind, args.entity_to_name)
    client.add_link(from_entity.uuid, to_entity.uuid)

def link_delete(args):
    client = KamaDatabaseClient(args)
    from_entity = client.get_entity(args.entity_from_kind, args.entity_from_name)
    to_entity = client.get_entity(args.entity_to_kind, args.entity_to_name)
    client.delete_link(from_entity.uuid, to_entity.uuid)


def _traverse_links(client, entity, parents=False, children=False, recursive=False, seen=None):
    results = []

    if seen is None:
        seen = set()

    if entity.uuid in seen:
        # Loop detected, short circuit
        return []

    seen.add(entity.uuid)

    if parents:
        results += [client.get_entity_by_uuid(x.from_entity.uuid) for x in entity.links_to]

    if children:
        results += [client.get_entity_by_uuid(x.to_entity.uuid) for x in entity.links_from]

    if recursive:
        for result in list(results):
            result = client.get_entity_by_uuid(result.uuid)
            results += _traverse_links(client, result, parents, children, recursive, seen)

    return results

def link_get(args):
    client = KamaDatabaseClient(args)
    entity = client.get_entity(args.entity_kind, args.entity_name)

    for result in _traverse_links(client, entity, args.parents, args.children, args.recursive):
        print('%s %s' % (result.kind, result.name))


def permission_add(args):
    client = KamaDatabaseClient(args)
    entity = client.get_entity(args.entity_kind, args.entity_name)
    role = client.get_entity('role', args.role_name)

    if args.permission_name.startswith('acl:'):
        permissions = DEFAULT_ACLS[args.permission_name.split(':', 1)[1]]
    else:
        permissions = [args.permission_name]

    for permission in permissions:
        client.add_permission(entity.uuid, role.uuid, permission)

def permission_delete(args):
    client = KamaDatabaseClient(args)
    entity = client.get_entity(args.entity_kind, args.entity_name)
    role = client.get_entity('role', args.role_name)

    if args.permission_name.startswith('acl:'):
        permissions = DEFAULT_ACLS[args.permission_name.split(':', 1)[1]]
    else:
        permissions = [args.permission_name]

    for permission in permissions:
        client.delete_permission(entity.uuid, role.uuid, permission)


def setup_arguments(parser):
    parser.add_argument('--server', default=kama.env.get(['KAMA_SERVER'], '127.0.0.1:8443'))
    parser.add_argument('--ca-cert', default=kama.env.get(['KAMA_CA_CERT'], 'ca-cert.pem'))
    parser.add_argument('--client-key', default=kama.env.get(['KAMA_CLIENT_KEY'], 'client.key'))
    parser.add_argument('--client-cert', default=kama.env.get(['KAMA_CLIENT_CERT'], 'client.cert'))
    sp = parser.add_subparsers()

    # kama dbclient entity ...
    entity_parser = sp.add_parser('entity')
    subparsers = entity_parser.add_subparsers()

    parser_entity_list = subparsers.add_parser('list')
    parser_entity_list.add_argument('--kind', default=None)
    parser_entity_list.set_defaults(func=list_entities)

    parser_entity_get = subparsers.add_parser('get')
    parser_entity_get.add_argument('kind')
    parser_entity_get.add_argument('name')
    parser_entity_get.set_defaults(func=get_entity)

    parser_entity_create = subparsers.add_parser('create')
    parser_entity_create.add_argument('kind')
    parser_entity_create.add_argument('name')
    parser_entity_create.add_argument('owner_role')
    parser_entity_create.set_defaults(func=create_entity)

    parser_entity_delete = subparsers.add_parser('delete')
    parser_entity_delete.add_argument('kind')
    parser_entity_delete.add_argument('name')
    parser_entity_delete.set_defaults(func=delete_entity)
    
    parser_entity_rename = subparsers.add_parser('rename')
    parser_entity_rename.add_argument('kind')
    parser_entity_rename.add_argument('old_name')
    parser_entity_rename.add_argument('new_name')
    parser_entity_rename.set_defaults(func=rename_entity)

    # kama dbclient attribute ...
    attribute_parser = sp.add_parser('attribute')
    subparsers = attribute_parser.add_subparsers()

    parser_attribute_add = subparsers.add_parser('add')
    parser_attribute_add.add_argument('entity_kind')
    parser_attribute_add.add_argument('entity_name')
    parser_attribute_add.add_argument('key')
    parser_attribute_add.add_argument('value')
    parser_attribute_add.set_defaults(func=attribute_add)

    parser_attribute_delete = subparsers.add_parser('delete')
    parser_attribute_delete.add_argument('entity_kind')
    parser_attribute_delete.add_argument('entity_name')
    parser_attribute_delete.add_argument('key')
    parser_attribute_delete.set_defaults(func=attribute_delete)

    parser_attribute_list = subparsers.add_parser('list')
    parser_attribute_list.add_argument('entity_kind')
    parser_attribute_list.add_argument('entity_name')
    parser_attribute_list.add_argument('--key', action='store', default=None)
    parser_attribute_list.set_defaults(func=attribute_list)

    # kama dbclient link ...
    link_parser = sp.add_parser('link')
    subparsers = link_parser.add_subparsers()
    parser_link_add = subparsers.add_parser('add')
    parser_link_add.add_argument('entity_from_kind')
    parser_link_add.add_argument('entity_from_name')
    parser_link_add.add_argument('entity_to_kind')
    parser_link_add.add_argument('entity_to_name')
    parser_link_add.set_defaults(func=link_add)

    parser_link_delete = subparsers.add_parser('delete')
    parser_link_delete.add_argument('entity_from_kind')
    parser_link_delete.add_argument('entity_from_name')
    parser_link_delete.add_argument('entity_to_kind')
    parser_link_delete.add_argument('entity_to_name')
    parser_link_delete.set_defaults(func=link_delete)

    parser_link_get = subparsers.add_parser('get')
    parser_link_get.add_argument('entity_kind')
    parser_link_get.add_argument('entity_name')
    parser_link_get.add_argument('--parents', action='store_true',
            help='List entities that link to this entity')
    parser_link_get.add_argument('--children', action='store_true',
            help='List entities that this entity links to')
    parser_link_get.add_argument('--recursive', action='store_true',
            help='Recursively traverse parents or children if either of those options is set')
    parser_link_get.set_defaults(func=link_get)

    # kama dbclient permission ...
    permission_parser = sp.add_parser('permission')
    subparsers = permission_parser.add_subparsers()
    parser_permission_add = subparsers.add_parser('add')
    parser_permission_add.add_argument('entity_kind')
    parser_permission_add.add_argument('entity_name')
    parser_permission_add.add_argument('role_name')
    parser_permission_add.add_argument('permission_name')
    parser_permission_add.set_defaults(func=permission_add)

    parser_permission_delete = subparsers.add_parser('delete')
    parser_permission_delete.add_argument('entity_kind')
    parser_permission_delete.add_argument('entity_name')
    parser_permission_delete.add_argument('role_name')
    parser_permission_delete.add_argument('permission_name')
    parser_permission_delete.set_defaults(func=permission_delete)


def main():
    parser = argparse.ArgumentParser()
    setup_arguments(parser)
    
    args = parser.parse_args()
    args.func(args)


if __name__ == '__main__':
    main()



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
import grpc

from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
from . import kama_pb2 as kama__pb2


class KamaDatabaseStub(object):
  # missing associated documentation comment in .proto file
  pass

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.ListEntities = channel.unary_stream(
        '/kama.KamaDatabase/ListEntities',
        request_serializer=kama__pb2.Entity.SerializeToString,
        response_deserializer=kama__pb2.Entity.FromString,
        )
    self.GetEntity = channel.unary_unary(
        '/kama.KamaDatabase/GetEntity',
        request_serializer=kama__pb2.Entity.SerializeToString,
        response_deserializer=kama__pb2.Entity.FromString,
        )
    self.CreateEntity = channel.unary_unary(
        '/kama.KamaDatabase/CreateEntity',
        request_serializer=kama__pb2.CreateEntityRequest.SerializeToString,
        response_deserializer=kama__pb2.Entity.FromString,
        )
    self.DeleteEntity = channel.unary_unary(
        '/kama.KamaDatabase/DeleteEntity',
        request_serializer=kama__pb2.Entity.SerializeToString,
        response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
        )
    self.UpdateEntity = channel.unary_unary(
        '/kama.KamaDatabase/UpdateEntity',
        request_serializer=kama__pb2.Entity.SerializeToString,
        response_deserializer=kama__pb2.Entity.FromString,
        )
    self.AddAttribute = channel.unary_unary(
        '/kama.KamaDatabase/AddAttribute',
        request_serializer=kama__pb2.Attribute.SerializeToString,
        response_deserializer=kama__pb2.Attribute.FromString,
        )
    self.DeleteAttributes = channel.unary_unary(
        '/kama.KamaDatabase/DeleteAttributes',
        request_serializer=kama__pb2.Attribute.SerializeToString,
        response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
        )
    self.AddLink = channel.unary_unary(
        '/kama.KamaDatabase/AddLink',
        request_serializer=kama__pb2.Link.SerializeToString,
        response_deserializer=kama__pb2.Link.FromString,
        )
    self.DeleteLink = channel.unary_unary(
        '/kama.KamaDatabase/DeleteLink',
        request_serializer=kama__pb2.Link.SerializeToString,
        response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
        )
    self.AddPermission = channel.unary_unary(
        '/kama.KamaDatabase/AddPermission',
        request_serializer=kama__pb2.Permission.SerializeToString,
        response_deserializer=kama__pb2.Permission.FromString,
        )
    self.DeletePermission = channel.unary_unary(
        '/kama.KamaDatabase/DeletePermission',
        request_serializer=kama__pb2.Permission.SerializeToString,
        response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
        )


class KamaDatabaseServicer(object):
  # missing associated documentation comment in .proto file
  pass

  def ListEntities(self, request, context):
    """ListEntities returns all Entities matching any of the non-null fields
    given in the request.
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetEntity(self, request, context):
    """GetEntity will try to find an entity with the given uuid, name, and
    kind, allowing wildcard searches for any of the fields that aren't set.
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def CreateEntity(self, request, context):
    """CreateEntity will create a new Entity with the given name and kind and
    return a new UUID
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def DeleteEntity(self, request, context):
    """DeleteEntity will delete the Entity with the given UUID. Other fields
    are ignored.
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def UpdateEntity(self, request, context):
    """UpdateEntity currently only supports changing the entity's name. It uses
    the uuid to determine what existing object is being updated. All other
    fields are ignored.
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def AddAttribute(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def DeleteAttributes(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def AddLink(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def DeleteLink(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def AddPermission(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def DeletePermission(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_KamaDatabaseServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'ListEntities': grpc.unary_stream_rpc_method_handler(
          servicer.ListEntities,
          request_deserializer=kama__pb2.Entity.FromString,
          response_serializer=kama__pb2.Entity.SerializeToString,
      ),
      'GetEntity': grpc.unary_unary_rpc_method_handler(
          servicer.GetEntity,
          request_deserializer=kama__pb2.Entity.FromString,
          response_serializer=kama__pb2.Entity.SerializeToString,
      ),
      'CreateEntity': grpc.unary_unary_rpc_method_handler(
          servicer.CreateEntity,
          request_deserializer=kama__pb2.CreateEntityRequest.FromString,
          response_serializer=kama__pb2.Entity.SerializeToString,
      ),
      'DeleteEntity': grpc.unary_unary_rpc_method_handler(
          servicer.DeleteEntity,
          request_deserializer=kama__pb2.Entity.FromString,
          response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
      ),
      'UpdateEntity': grpc.unary_unary_rpc_method_handler(
          servicer.UpdateEntity,
          request_deserializer=kama__pb2.Entity.FromString,
          response_serializer=kama__pb2.Entity.SerializeToString,
      ),
      'AddAttribute': grpc.unary_unary_rpc_method_handler(
          servicer.AddAttribute,
          request_deserializer=kama__pb2.Attribute.FromString,
          response_serializer=kama__pb2.Attribute.SerializeToString,
      ),
      'DeleteAttributes': grpc.unary_unary_rpc_method_handler(
          servicer.DeleteAttributes,
          request_deserializer=kama__pb2.Attribute.FromString,
          response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
      ),
      'AddLink': grpc.unary_unary_rpc_method_handler(
          servicer.AddLink,
          request_deserializer=kama__pb2.Link.FromString,
          response_serializer=kama__pb2.Link.SerializeToString,
      ),
      'DeleteLink': grpc.unary_unary_rpc_method_handler(
          servicer.DeleteLink,
          request_deserializer=kama__pb2.Link.FromString,
          response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
      ),
      'AddPermission': grpc.unary_unary_rpc_method_handler(
          servicer.AddPermission,
          request_deserializer=kama__pb2.Permission.FromString,
          response_serializer=kama__pb2.Permission.SerializeToString,
      ),
      'DeletePermission': grpc.unary_unary_rpc_method_handler(
          servicer.DeletePermission,
          request_deserializer=kama__pb2.Permission.FromString,
          response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'kama.KamaDatabase', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))



import kama.database


class AuthenticationException(Exception):
    pass


def name_from_grpc_context(grpc_context):
    cname = grpc_context.peer_identities()[0]
    cname = cname.split('.')
    name, kind = cname[:2]

    return (kind, name)


def user_from_grpc_context(grpc_context):
    kind, name = name_from_grpc_context(grpc_context)
    user = kama.database.Entity.get_by_name(kind, name)
    if user is None:
        raise AuthenticationException('%s/%s does not exist!' % (kind, name))
    return user


def get_request_context(grpc_context):
    user = user_from_grpc_context(grpc_context)
    return kama.database.RequestContext(user)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: kama.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='kama.proto',
  package='kama',
  syntax='proto3',
  serialized_pb=_b('\n\nkama.proto\x12\x04kama\x1a\x1bgoogle/protobuf/empty.proto\"\xbc\x01\n\x06\x45ntity\x12\x0c\n\x04uuid\x18\x01 \x01(\x0c\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x0c\n\x04kind\x18\x03 \x01(\t\x12#\n\nattributes\x18\x04 \x03(\x0b\x32\x0f.kama.Attribute\x12\x1e\n\nlinks_from\x18\x05 \x03(\x0b\x32\n.kama.Link\x12\x1c\n\x08links_to\x18\x06 \x03(\x0b\x32\n.kama.Link\x12%\n\x0bpermissions\x18\x07 \x03(\x0b\x32\x10.kama.Permission\"S\n\tAttribute\x12\x0c\n\x04uuid\x18\x01 \x01(\x0c\x12\x1c\n\x06\x65ntity\x18\x02 \x01(\x0b\x32\x0c.kama.Entity\x12\x0b\n\x03key\x18\x03 \x01(\t\x12\r\n\x05value\x18\x04 \x01(\x0c\"X\n\x04Link\x12\x0c\n\x04uuid\x18\x01 \x01(\x0c\x12!\n\x0b\x66rom_entity\x18\x02 \x01(\x0b\x32\x0c.kama.Entity\x12\x1f\n\tto_entity\x18\x03 \x01(\x0b\x32\x0c.kama.Entity\"b\n\nPermission\x12\x0c\n\x04uuid\x18\x01 \x01(\x0c\x12\x1a\n\x04role\x18\x02 \x01(\x0b\x32\x0c.kama.Entity\x12\x1c\n\x06\x65ntity\x18\x03 \x01(\x0b\x32\x0c.kama.Entity\x12\x0c\n\x04name\x18\x04 \x01(\t\"U\n\x13\x43reateEntityRequest\x12\x1c\n\x06\x65ntity\x18\x01 \x01(\x0b\x32\x0c.kama.Entity\x12 \n\nowner_role\x18\x02 \x01(\x0b\x32\x0c.kama.Entity2\xb7\x04\n\x0cKamaDatabase\x12,\n\x0cListEntities\x12\x0c.kama.Entity\x1a\x0c.kama.Entity0\x01\x12\'\n\tGetEntity\x12\x0c.kama.Entity\x1a\x0c.kama.Entity\x12\x37\n\x0c\x43reateEntity\x12\x19.kama.CreateEntityRequest\x1a\x0c.kama.Entity\x12\x34\n\x0c\x44\x65leteEntity\x12\x0c.kama.Entity\x1a\x16.google.protobuf.Empty\x12*\n\x0cUpdateEntity\x12\x0c.kama.Entity\x1a\x0c.kama.Entity\x12\x30\n\x0c\x41\x64\x64\x41ttribute\x12\x0f.kama.Attribute\x1a\x0f.kama.Attribute\x12;\n\x10\x44\x65leteAttributes\x12\x0f.kama.Attribute\x1a\x16.google.protobuf.Empty\x12!\n\x07\x41\x64\x64Link\x12\n.kama.Link\x1a\n.kama.Link\x12\x30\n\nDeleteLink\x12\n.kama.Link\x1a\x16.google.protobuf.Empty\x12\x33\n\rAddPermission\x12\x10.kama.Permission\x1a\x10.kama.Permission\x12<\n\x10\x44\x65letePermission\x12\x10.kama.Permission\x1a\x16.google.protobuf.Emptyb\x06proto3')
  ,
  dependencies=[google_dot_protobuf_dot_empty__pb2.DESCRIPTOR,])




_ENTITY = _descriptor.Descriptor(
  name='Entity',
  full_name='kama.Entity',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='uuid', full_name='kama.Entity.uuid', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='name', full_name='kama.Entity.name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='kind', full_name='kama.Entity.kind', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='attributes', full_name='kama.Entity.attributes', index=3,
      number=4, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='links_from', full_name='kama.Entity.links_from', index=4,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='links_to', full_name='kama.Entity.links_to', index=5,
      number=6, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='permissions', full_name='kama.Entity.permissions', index=6,
      number=7, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=50,
  serialized_end=238,
)


_ATTRIBUTE = _descriptor.Descriptor(
  name='Attribute',
  full_name='kama.Attribute',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='uuid', full_name='kama.Attribute.uuid', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='entity', full_name='kama.Attribute.entity', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='key', full_name='kama.Attribute.key', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='value', full_name='kama.Attribute.value', index=3,
      number=4, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=240,
  serialized_end=323,
)


_LINK = _descriptor.Descriptor(
  name='Link',
  full_name='kama.Link',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='uuid', full_name='kama.Link.uuid', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='from_entity', full_name='kama.Link.from_entity', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='to_entity', full_name='kama.Link.to_entity', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=325,
  serialized_end=413,
)


_PERMISSION = _descriptor.Descriptor(
  name='Permission',
  full_name='kama.Permission',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='uuid', full_name='kama.Permission.uuid', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='role', full_name='kama.Permission.role', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='entity', full_name='kama.Permission.entity', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='name', full_name='kama.Permission.name', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=415,
  serialized_end=513,
)


_CREATEENTITYREQUEST = _descriptor.Descriptor(
  name='CreateEntityRequest',
  full_name='kama.CreateEntityRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='entity', full_name='kama.CreateEntityRequest.entity', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='owner_role', full_name='kama.CreateEntityRequest.owner_role', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=515,
  serialized_end=600,
)

_ENTITY.fields_by_name['attributes'].message_type = _ATTRIBUTE
_ENTITY.fields_by_name['links_from'].message_type = _LINK
_ENTITY.fields_by_name['links_to'].message_type = _LINK
_ENTITY.fields_by_name['permissions'].message_type = _PERMISSION
_ATTRIBUTE.fields_by_name['entity'].message_type = _ENTITY
_LINK.fields_by_name['from_entity'].message_type = _ENTITY
_LINK.fields_by_name['to_entity'].message_type = _ENTITY
_PERMISSION.fields_by_name['role'].message_type = _ENTITY
_PERMISSION.fields_by_name['entity'].message_type = _ENTITY
_CREATEENTITYREQUEST.fields_by_name['entity'].message_type = _ENTITY
_CREATEENTITYREQUEST.fields_by_name['owner_role'].message_type = _ENTITY
DESCRIPTOR.message_types_by_name['Entity'] = _ENTITY
DESCRIPTOR.message_types_by_name['Attribute'] = _ATTRIBUTE
DESCRIPTOR.message_types_by_name['Link'] = _LINK
DESCRIPTOR.message_types_by_name['Permission'] = _PERMISSION
DESCRIPTOR.message_types_by_name['CreateEntityRequest'] = _CREATEENTITYREQUEST
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Entity = _reflection.GeneratedProtocolMessageType('Entity', (_message.Message,), dict(
  DESCRIPTOR = _ENTITY,
  __module__ = 'kama_pb2'
  # @@protoc_insertion_point(class_scope:kama.Entity)
  ))
_sym_db.RegisterMessage(Entity)

Attribute = _reflection.GeneratedProtocolMessageType('Attribute', (_message.Message,), dict(
  DESCRIPTOR = _ATTRIBUTE,
  __module__ = 'kama_pb2'
  # @@protoc_insertion_point(class_scope:kama.Attribute)
  ))
_sym_db.RegisterMessage(Attribute)

Link = _reflection.GeneratedProtocolMessageType('Link', (_message.Message,), dict(
  DESCRIPTOR = _LINK,
  __module__ = 'kama_pb2'
  # @@protoc_insertion_point(class_scope:kama.Link)
  ))
_sym_db.RegisterMessage(Link)

Permission = _reflection.GeneratedProtocolMessageType('Permission', (_message.Message,), dict(
  DESCRIPTOR = _PERMISSION,
  __module__ = 'kama_pb2'
  # @@protoc_insertion_point(class_scope:kama.Permission)
  ))
_sym_db.RegisterMessage(Permission)

CreateEntityRequest = _reflection.GeneratedProtocolMessageType('CreateEntityRequest', (_message.Message,), dict(
  DESCRIPTOR = _CREATEENTITYREQUEST,
  __module__ = 'kama_pb2'
  # @@protoc_insertion_point(class_scope:kama.CreateEntityRequest)
  ))
_sym_db.RegisterMessage(CreateEntityRequest)


try:
  # THESE ELEMENTS WILL BE DEPRECATED.
  # Please use the generated *_pb2_grpc.py files instead.
  import grpc
  from grpc.beta import implementations as beta_implementations
  from grpc.beta import interfaces as beta_interfaces
  from grpc.framework.common import cardinality
  from grpc.framework.interfaces.face import utilities as face_utilities


  class KamaDatabaseStub(object):
    # missing associated documentation comment in .proto file
    pass

    def __init__(self, channel):
      """Constructor.

      Args:
        channel: A grpc.Channel.
      """
      self.ListEntities = channel.unary_stream(
          '/kama.KamaDatabase/ListEntities',
          request_serializer=Entity.SerializeToString,
          response_deserializer=Entity.FromString,
          )
      self.GetEntity = channel.unary_unary(
          '/kama.KamaDatabase/GetEntity',
          request_serializer=Entity.SerializeToString,
          response_deserializer=Entity.FromString,
          )
      self.CreateEntity = channel.unary_unary(
          '/kama.KamaDatabase/CreateEntity',
          request_serializer=CreateEntityRequest.SerializeToString,
          response_deserializer=Entity.FromString,
          )
      self.DeleteEntity = channel.unary_unary(
          '/kama.KamaDatabase/DeleteEntity',
          request_serializer=Entity.SerializeToString,
          response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
          )
      self.UpdateEntity = channel.unary_unary(
          '/kama.KamaDatabase/UpdateEntity',
          request_serializer=Entity.SerializeToString,
          response_deserializer=Entity.FromString,
          )
      self.AddAttribute = channel.unary_unary(
          '/kama.KamaDatabase/AddAttribute',
          request_serializer=Attribute.SerializeToString,
          response_deserializer=Attribute.FromString,
          )
      self.DeleteAttributes = channel.unary_unary(
          '/kama.KamaDatabase/DeleteAttributes',
          request_serializer=Attribute.SerializeToString,
          response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
          )
      self.AddLink = channel.unary_unary(
          '/kama.KamaDatabase/AddLink',
          request_serializer=Link.SerializeToString,
          response_deserializer=Link.FromString,
          )
      self.DeleteLink = channel.unary_unary(
          '/kama.KamaDatabase/DeleteLink',
          request_serializer=Link.SerializeToString,
          response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
          )
      self.AddPermission = channel.unary_unary(
          '/kama.KamaDatabase/AddPermission',
          request_serializer=Permission.SerializeToString,
          response_deserializer=Permission.FromString,
          )
      self.DeletePermission = channel.unary_unary(
          '/kama.KamaDatabase/DeletePermission',
          request_serializer=Permission.SerializeToString,
          response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
          )


  class KamaDatabaseServicer(object):
    # missing associated documentation comment in .proto file
    pass

    def ListEntities(self, request, context):
      """ListEntities returns all Entities matching any of the non-null fields
      given in the request.
      """
      context.set_code(grpc.StatusCode.UNIMPLEMENTED)
      context.set_details('Method not implemented!')
      raise NotImplementedError('Method not implemented!')

    def GetEntity(self, request, context):
      """GetEntity will try to find an entity with the given uuid, name, and
      kind, allowing wildcard searches for any of the fields that aren't set.
      """
      context.set_code(grpc.StatusCode.UNIMPLEMENTED)
      context.set_details('Method not implemented!')
      raise NotImplementedError('Method not implemented!')

    def CreateEntity(self, request, context):
      """CreateEntity will create a new Entity with the given name and kind and
      return a new UUID
      """
      context.set_code(grpc.StatusCode.UNIMPLEMENTED)
      context.set_details('Method not implemented!')
      raise NotImplementedError('Method not implemented!')

    def DeleteEntity(self, request, context):
      """DeleteEntity will delete the Entity with the given UUID. Other fields
      are ignored.
      """
      context.set_code(grpc.StatusCode.UNIMPLEMENTED)
      context.set_details('Method not implemented!')
      raise NotImplementedError('Method not implemented!')

    def UpdateEntity(self, request, context):
      """UpdateEntity currently only supports changing the entity's name. It uses
      the uuid to determine what existing object is being updated. All other
      fields are ignored.
      """
      context.set_code(grpc.StatusCode.UNIMPLEMENTED)
      context.set_details('Method not implemented!')
      raise NotImplementedError('Method not implemented!')

    def AddAttribute(self, request, context):
      # missing associated documentation comment in .proto file
      pass
      context.set_code(grpc.StatusCode.UNIMPLEMENTED)
      context.set_details('Method not implemented!')
      raise NotImplementedError('Method not implemented!')

    def DeleteAttributes(self, request, context):
      # missing associated documentation comment in .proto file
      pass
      context.set_code(grpc.StatusCode.UNIMPLEMENTED)
      context.set_details('Method not implemented!')
      raise NotImplementedError('Method not implemented!')

    def AddLink(self, request, context):
      # missing associated documentation comment in .proto file
      pass
      context.set_code(grpc.StatusCode.UNIMPLEMENTED)
      context.set_details('Method not implemented!')
      raise NotImplementedError('Method not implemented!')

    def DeleteLink(self, request, context):
      # missing associated documentation comment in .proto file
      pass
      context.set_code(grpc.StatusCode.UNIMPLEMENTED)
      context.set_details('Method not implemented!')
      raise NotImplementedError('Method not implemented!')

    def AddPermission(self, request, context):
      # missing associated documentation comment in .proto file
      pass
      context.set_code(grpc.StatusCode.UNIMPLEMENTED)
      context.set_details('Method not implemented!')
      raise NotImplementedError('Method not implemented!')

    def DeletePermission(self, request, context):
      # missing associated documentation comment in .proto file
      pass
      context.set_code(grpc.StatusCode.UNIMPLEMENTED)
      context.set_details('Method not implemented!')
      raise NotImplementedError('Method not implemented!')


  def add_KamaDatabaseServicer_to_server(servicer, server):
    rpc_method_handlers = {
        'ListEntities': grpc.unary_stream_rpc_method_handler(
            servicer.ListEntities,
            request_deserializer=Entity.FromString,
            response_serializer=Entity.SerializeToString,
        ),
        'GetEntity': grpc.unary_unary_rpc_method_handler(
            servicer.GetEntity,
            request_deserializer=Entity.FromString,
            response_serializer=Entity.SerializeToString,
        ),
        'CreateEntity': grpc.unary_unary_rpc_method_handler(
            servicer.CreateEntity,
            request_deserializer=CreateEntityRequest.FromString,
            response_serializer=Entity.SerializeToString,
        ),
        'DeleteEntity': grpc.unary_unary_rpc_method_handler(
            servicer.DeleteEntity,
            request_deserializer=Entity.FromString,
            response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
        ),
        'UpdateEntity': grpc.unary_unary_rpc_method_handler(
            servicer.UpdateEntity,
            request_deserializer=Entity.FromString,
            response_serializer=Entity.SerializeToString,
        ),
        'AddAttribute': grpc.unary_unary_rpc_method_handler(
            servicer.AddAttribute,
            request_deserializer=Attribute.FromString,
            response_serializer=Attribute.SerializeToString,
        ),
        'DeleteAttributes': grpc.unary_unary_rpc_method_handler(
            servicer.DeleteAttributes,
            request_deserializer=Attribute.FromString,
            response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
        ),
        'AddLink': grpc.unary_unary_rpc_method_handler(
            servicer.AddLink,
            request_deserializer=Link.FromString,
            response_serializer=Link.SerializeToString,
        ),
        'DeleteLink': grpc.unary_unary_rpc_method_handler(
            servicer.DeleteLink,
            request_deserializer=Link.FromString,
            response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
        ),
        'AddPermission': grpc.unary_unary_rpc_method_handler(
            servicer.AddPermission,
            request_deserializer=Permission.FromString,
            response_serializer=Permission.SerializeToString,
        ),
        'DeletePermission': grpc.unary_unary_rpc_method_handler(
            servicer.DeletePermission,
            request_deserializer=Permission.FromString,
            response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
        ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
        'kama.KamaDatabase', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


  class BetaKamaDatabaseServicer(object):
    """The Beta API is deprecated for 0.15.0 and later.

    It is recommended to use the GA API (classes and functions in this
    file not marked beta) for all further purposes. This class was generated
    only to ease transition from grpcio<0.15.0 to grpcio>=0.15.0."""
    # missing associated documentation comment in .proto file
    pass
    def ListEntities(self, request, context):
      """ListEntities returns all Entities matching any of the non-null fields
      given in the request.
      """
      context.code(beta_interfaces.StatusCode.UNIMPLEMENTED)
    def GetEntity(self, request, context):
      """GetEntity will try to find an entity with the given uuid, name, and
      kind, allowing wildcard searches for any of the fields that aren't set.
      """
      context.code(beta_interfaces.StatusCode.UNIMPLEMENTED)
    def CreateEntity(self, request, context):
      """CreateEntity will create a new Entity with the given name and kind and
      return a new UUID
      """
      context.code(beta_interfaces.StatusCode.UNIMPLEMENTED)
    def DeleteEntity(self, request, context):
      """DeleteEntity will delete the Entity with the given UUID. Other fields
      are ignored.
      """
      context.code(beta_interfaces.StatusCode.UNIMPLEMENTED)
    def UpdateEntity(self, request, context):
      """UpdateEntity currently only supports changing the entity's name. It uses
      the uuid to determine what existing object is being updated. All other
      fields are ignored.
      """
      context.code(beta_interfaces.StatusCode.UNIMPLEMENTED)
    def AddAttribute(self, request, context):
      # missing associated documentation comment in .proto file
      pass
      context.code(beta_interfaces.StatusCode.UNIMPLEMENTED)
    def DeleteAttributes(self, request, context):
      # missing associated documentation comment in .proto file
      pass
      context.code(beta_interfaces.StatusCode.UNIMPLEMENTED)
    def AddLink(self, request, context):
      # missing associated documentation comment in .proto file
      pass
      context.code(beta_interfaces.StatusCode.UNIMPLEMENTED)
    def DeleteLink(self, request, context):
      # missing associated documentation comment in .proto file
      pass
      context.code(beta_interfaces.StatusCode.UNIMPLEMENTED)
    def AddPermission(self, request, context):
      # missing associated documentation comment in .proto file
      pass
      context.code(beta_interfaces.StatusCode.UNIMPLEMENTED)
    def DeletePermission(self, request, context):
      # missing associated documentation comment in .proto file
      pass
      context.code(beta_interfaces.StatusCode.UNIMPLEMENTED)


  class BetaKamaDatabaseStub(object):
    """The Beta API is deprecated for 0.15.0 and later.

    It is recommended to use the GA API (classes and functions in this
    file not marked beta) for all further purposes. This class was generated
    only to ease transition from grpcio<0.15.0 to grpcio>=0.15.0."""
    # missing associated documentation comment in .proto file
    pass
    def ListEntities(self, request, timeout, metadata=None, with_call=False, protocol_options=None):
      """ListEntities returns all Entities matching any of the non-null fields
      given in the request.
      """
      raise NotImplementedError()
    def GetEntity(self, request, timeout, metadata=None, with_call=False, protocol_options=None):
      """GetEntity will try to find an entity with the given uuid, name, and
      kind, allowing wildcard searches for any of the fields that aren't set.
      """
      raise NotImplementedError()
    GetEntity.future = None
    def CreateEntity(self, request, timeout, metadata=None, with_call=False, protocol_options=None):
      """CreateEntity will create a new Entity with the given name and kind and
      return a new UUID
      """
      raise NotImplementedError()
    CreateEntity.future = None
    def DeleteEntity(self, request, timeout, metadata=None, with_call=False, protocol_options=None):
      """DeleteEntity will delete the Entity with the given UUID. Other fields
      are ignored.
      """
      raise NotImplementedError()
    DeleteEntity.future = None
    def UpdateEntity(self, request, timeout, metadata=None, with_call=False, protocol_options=None):
      """UpdateEntity currently only supports changing the entity's name. It uses
      the uuid to determine what existing object is being updated. All other
      fields are ignored.
      """
      raise NotImplementedError()
    UpdateEntity.future = None
    def AddAttribute(self, request, timeout, metadata=None, with_call=False, protocol_options=None):
      # missing associated documentation comment in .proto file
      pass
      raise NotImplementedError()
    AddAttribute.future = None
    def DeleteAttributes(self, request, timeout, metadata=None, with_call=False, protocol_options=None):
      # missing associated documentation comment in .proto file
      pass
      raise NotImplementedError()
    DeleteAttributes.future = None
    def AddLink(self, request, timeout, metadata=None, with_call=False, protocol_options=None):
      # missing associated documentation comment in .proto file
      pass
      raise NotImplementedError()
    AddLink.future = None
    def DeleteLink(self, request, timeout, metadata=None, with_call=False, protocol_options=None):
      # missing associated documentation comment in .proto file
      pass
      raise NotImplementedError()
    DeleteLink.future = None
    def AddPermission(self, request, timeout, metadata=None, with_call=False, protocol_options=None):
      # missing associated documentation comment in .proto file
      pass
      raise NotImplementedError()
    AddPermission.future = None
    def DeletePermission(self, request, timeout, metadata=None, with_call=False, protocol_options=None):
      # missing associated documentation comment in .proto file
      pass
      raise NotImplementedError()
    DeletePermission.future = None


  def beta_create_KamaDatabase_server(servicer, pool=None, pool_size=None, default_timeout=None, maximum_timeout=None):
    """The Beta API is deprecated for 0.15.0 and later.

    It is recommended to use the GA API (classes and functions in this
    file not marked beta) for all further purposes. This function was
    generated only to ease transition from grpcio<0.15.0 to grpcio>=0.15.0"""
    request_deserializers = {
      ('kama.KamaDatabase', 'AddAttribute'): Attribute.FromString,
      ('kama.KamaDatabase', 'AddLink'): Link.FromString,
      ('kama.KamaDatabase', 'AddPermission'): Permission.FromString,
      ('kama.KamaDatabase', 'CreateEntity'): CreateEntityRequest.FromString,
      ('kama.KamaDatabase', 'DeleteAttributes'): Attribute.FromString,
      ('kama.KamaDatabase', 'DeleteEntity'): Entity.FromString,
      ('kama.KamaDatabase', 'DeleteLink'): Link.FromString,
      ('kama.KamaDatabase', 'DeletePermission'): Permission.FromString,
      ('kama.KamaDatabase', 'GetEntity'): Entity.FromString,
      ('kama.KamaDatabase', 'ListEntities'): Entity.FromString,
      ('kama.KamaDatabase', 'UpdateEntity'): Entity.FromString,
    }
    response_serializers = {
      ('kama.KamaDatabase', 'AddAttribute'): Attribute.SerializeToString,
      ('kama.KamaDatabase', 'AddLink'): Link.SerializeToString,
      ('kama.KamaDatabase', 'AddPermission'): Permission.SerializeToString,
      ('kama.KamaDatabase', 'CreateEntity'): Entity.SerializeToString,
      ('kama.KamaDatabase', 'DeleteAttributes'): google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
      ('kama.KamaDatabase', 'DeleteEntity'): google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
      ('kama.KamaDatabase', 'DeleteLink'): google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
      ('kama.KamaDatabase', 'DeletePermission'): google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
      ('kama.KamaDatabase', 'GetEntity'): Entity.SerializeToString,
      ('kama.KamaDatabase', 'ListEntities'): Entity.SerializeToString,
      ('kama.KamaDatabase', 'UpdateEntity'): Entity.SerializeToString,
    }
    method_implementations = {
      ('kama.KamaDatabase', 'AddAttribute'): face_utilities.unary_unary_inline(servicer.AddAttribute),
      ('kama.KamaDatabase', 'AddLink'): face_utilities.unary_unary_inline(servicer.AddLink),
      ('kama.KamaDatabase', 'AddPermission'): face_utilities.unary_unary_inline(servicer.AddPermission),
      ('kama.KamaDatabase', 'CreateEntity'): face_utilities.unary_unary_inline(servicer.CreateEntity),
      ('kama.KamaDatabase', 'DeleteAttributes'): face_utilities.unary_unary_inline(servicer.DeleteAttributes),
      ('kama.KamaDatabase', 'DeleteEntity'): face_utilities.unary_unary_inline(servicer.DeleteEntity),
      ('kama.KamaDatabase', 'DeleteLink'): face_utilities.unary_unary_inline(servicer.DeleteLink),
      ('kama.KamaDatabase', 'DeletePermission'): face_utilities.unary_unary_inline(servicer.DeletePermission),
      ('kama.KamaDatabase', 'GetEntity'): face_utilities.unary_unary_inline(servicer.GetEntity),
      ('kama.KamaDatabase', 'ListEntities'): face_utilities.unary_stream_inline(servicer.ListEntities),
      ('kama.KamaDatabase', 'UpdateEntity'): face_utilities.unary_unary_inline(servicer.UpdateEntity),
    }
    server_options = beta_implementations.server_options(request_deserializers=request_deserializers, response_serializers=response_serializers, thread_pool=pool, thread_pool_size=pool_size, default_timeout=default_timeout, maximum_timeout=maximum_timeout)
    return beta_implementations.server(method_implementations, options=server_options)


  def beta_create_KamaDatabase_stub(channel, host=None, metadata_transformer=None, pool=None, pool_size=None):
    """The Beta API is deprecated for 0.15.0 and later.

    It is recommended to use the GA API (classes and functions in this
    file not marked beta) for all further purposes. This function was
    generated only to ease transition from grpcio<0.15.0 to grpcio>=0.15.0"""
    request_serializers = {
      ('kama.KamaDatabase', 'AddAttribute'): Attribute.SerializeToString,
      ('kama.KamaDatabase', 'AddLink'): Link.SerializeToString,
      ('kama.KamaDatabase', 'AddPermission'): Permission.SerializeToString,
      ('kama.KamaDatabase', 'CreateEntity'): CreateEntityRequest.SerializeToString,
      ('kama.KamaDatabase', 'DeleteAttributes'): Attribute.SerializeToString,
      ('kama.KamaDatabase', 'DeleteEntity'): Entity.SerializeToString,
      ('kama.KamaDatabase', 'DeleteLink'): Link.SerializeToString,
      ('kama.KamaDatabase', 'DeletePermission'): Permission.SerializeToString,
      ('kama.KamaDatabase', 'GetEntity'): Entity.SerializeToString,
      ('kama.KamaDatabase', 'ListEntities'): Entity.SerializeToString,
      ('kama.KamaDatabase', 'UpdateEntity'): Entity.SerializeToString,
    }
    response_deserializers = {
      ('kama.KamaDatabase', 'AddAttribute'): Attribute.FromString,
      ('kama.KamaDatabase', 'AddLink'): Link.FromString,
      ('kama.KamaDatabase', 'AddPermission'): Permission.FromString,
      ('kama.KamaDatabase', 'CreateEntity'): Entity.FromString,
      ('kama.KamaDatabase', 'DeleteAttributes'): google_dot_protobuf_dot_empty__pb2.Empty.FromString,
      ('kama.KamaDatabase', 'DeleteEntity'): google_dot_protobuf_dot_empty__pb2.Empty.FromString,
      ('kama.KamaDatabase', 'DeleteLink'): google_dot_protobuf_dot_empty__pb2.Empty.FromString,
      ('kama.KamaDatabase', 'DeletePermission'): google_dot_protobuf_dot_empty__pb2.Empty.FromString,
      ('kama.KamaDatabase', 'GetEntity'): Entity.FromString,
      ('kama.KamaDatabase', 'ListEntities'): Entity.FromString,
      ('kama.KamaDatabase', 'UpdateEntity'): Entity.FromString,
    }
    cardinalities = {
      'AddAttribute': cardinality.Cardinality.UNARY_UNARY,
      'AddLink': cardinality.Cardinality.UNARY_UNARY,
      'AddPermission': cardinality.Cardinality.UNARY_UNARY,
      'CreateEntity': cardinality.Cardinality.UNARY_UNARY,
      'DeleteAttributes': cardinality.Cardinality.UNARY_UNARY,
      'DeleteEntity': cardinality.Cardinality.UNARY_UNARY,
      'DeleteLink': cardinality.Cardinality.UNARY_UNARY,
      'DeletePermission': cardinality.Cardinality.UNARY_UNARY,
      'GetEntity': cardinality.Cardinality.UNARY_UNARY,
      'ListEntities': cardinality.Cardinality.UNARY_STREAM,
      'UpdateEntity': cardinality.Cardinality.UNARY_UNARY,
    }
    stub_options = beta_implementations.stub_options(host=host, metadata_transformer=metadata_transformer, request_serializers=request_serializers, response_deserializers=response_deserializers, thread_pool=pool, thread_pool_size=pool_size)
    return beta_implementations.dynamic_stub(channel, 'kama.KamaDatabase', cardinalities, options=stub_options)
except ImportError:
  pass
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
import MySQLdb
import MySQLdb.cursors
import threading
import functools
import uuid
import time
import sys
import os

import kama.acl
import kama.env
import kama.log


CACHE_TTL = 60.0


log = kama.log.get_logger('kama.database')
audit_log = kama.log.get_logger('kama.audit')
tlocal = threading.local()


class KamaException(Exception):
    pass


class EntityNotFoundException(KamaException):
    pass


class AttributeNotFoundException(KamaException):
    pass


class LinkNotFoundException(KamaException):
    pass


class PermissionNotFoundException(KamaException):
    pass


class PermissionDeniedException(KamaException):
    pass


def generate_uuid():
    # This is not RFC4122 UUID strictly, it's a re-ordered version of
    # UUID putting the node ID first. This makes it gentler on MySQL
    # when used as a primary key as newly generated keys will usually
    # come after older ones, so writes are more likely to be appends.
    #
    # Further, we ensure that it's encoded as an upper-case hex
    # string, because MySQL's HEX() function uses that and it makes
    # equality comparisons easier this way.
    version = 15    # definitely out of spec.
    u = uuid.uuid1()
    u = (version << 122) | (u.node << 74) | (u.time << 14) | u.clock_seq
    return '%032X' % u


class LoggingCursor(MySQLdb.cursors.Cursor):
    def execute(self, query, args=None):
        log.debug('execute: %r %% %r', query, args)
        return MySQLdb.cursors.Cursor.execute(self, query, args)


class DatabaseContext(object):
    def __init__(self):
        self.database = None

    def __enter__(self, retries=0):
        try:
            if self.database is None:
                params = {
                    'host': kama.env.get(['MYSQL_SERVICE_HOST', 'MYSQL_PORT_3306_TCP_ADDR'], '127.0.0.1'),
                    'port': int(kama.env.get(['MYSQL_SERVICE_PORT', 'MYSQL_PORT_3306_TCP_PORT'], 3306)),
                    'user': kama.env.get(['MYSQL_SERVICE_USER'], 'kama'),
                    'db': kama.env.get(['MYSQL_SERVICE_DATABASE'], 'kama'),
                    'connect_timeout': int(kama.env.get(['MYSQL_CONNECT_TIMEOUT'], 5)),
                }
                log.debug('Connecting to database with params: %r', params)

                params['passwd'] = kama.env.get(['MYSQL_SERVICE_PASSWORD'])
                if params['passwd'] is None:
                    del params['passwd']

                self.database = MySQLdb.connect(**params)

            cursor = self.database.cursor(cursorclass=LoggingCursor)
            cursor.execute('SELECT VERSION()')
            version = cursor.fetchall()[0][0]
            log.debug('Connected to database %s' % version)
            #cursor.execute('SET NAMES utf8mb4')
            #cursor.execute('SET CHARACTER SET utf8mb4')
            #cursor.execute('SET character_set_connection=utf8mb4')
            return cursor
        except Exception as e:
            log.warning('Database error: %s', str(e))
            if self.database is not None:
                self.database.close()
                self.database = None
            retries += 1
            if retries > 2:
                log.error('Max retries hit, hard fail.')
                raise
            log.info('Retrying database connection, retry #%i' % retries)
            return self.__enter__(retries=retries)

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.database.__exit__(exc_type, exc_value, exc_traceback)
        if exc_type is not None:
            try:
                log.debug('Closing database connection after exception was handled: %s', exc_type)
                self.database.close()
                self.database = None
            except: pass
        else:
            log.debug('COMMIT')
            self.database.commit()


def get_database_context():
    if not hasattr(tlocal, 'database_context'):
        tlocal.database_context = DatabaseContext()
    return tlocal.database_context


def schema_init():
    with get_database_context() as cursor:
        cursor.execute('''
CREATE TABLE `entities` (
    `entity_uuid` BINARY(16) NOT NULL PRIMARY KEY,
    `kind` TEXT NOT NULL,
    `name` TEXT NOT NULL
)''')

        cursor.execute('''
CREATE TABLE `attributes` (
    `attribute_uuid` BINARY(16) NOT NULL PRIMARY KEY,
    `entity_uuid` BINARY(16) NOT NULL,
    `key` TEXT NOT NULL,
    `value` BLOB NOT NULL
)''')

        cursor.execute('''
CREATE TABLE `links` (
    `link_uuid` BINARY(16) NOT NULL PRIMARY KEY,
    `from_uuid` BINARY(16) NOT NULL,
    `to_uuid` BINARY(16) NOT NULL
)''')

        cursor.execute('''
CREATE TABLE `permissions` (
    `permission_uuid` BINARY(16) NOT NULL PRIMARY KEY,
    `role_uuid` BINARY(16) NOT NULL,
    `entity_uuid` BINARY(16) NOT NULL,
    `name` TEXT NOT NULL
)''')

    with get_database_context() as cursor:
        cursor.execute('CREATE INDEX `idx_entities_kind` ON `entities` (`kind`(3))')
        cursor.execute('CREATE INDEX `idx_entities_name` ON `entities` (`name`(3))')
        cursor.execute('CREATE INDEX `idx_attributes_key` on `attributes` (`key`(3))')
        cursor.execute('CREATE INDEX `idx_links_to_uuid` ON `links` (`to_uuid`)')
        cursor.execute('CREATE INDEX `idx_links_from_uuid` ON `links` (`from_uuid`)')

    with get_database_context() as cursor:
        # We have to manually create the first role and user because
        # new Entities require an owner role to be created.
        role_uuid = generate_uuid()
        user_uuid = generate_uuid()
        link_uuid = generate_uuid()
        cursor.execute('INSERT INTO `entities` (entity_uuid, kind, name) VALUES(UNHEX(%s), %s, %s)', (role_uuid, 'role', 'root'))
        cursor.execute('INSERT INTO `entities` (entity_uuid, kind, name) VALUES(UNHEX(%s), %s, %s)', (user_uuid, 'user', 'root'))
        cursor.execute('INSERT INTO `links` (link_uuid, from_uuid, to_uuid) VALUES(UNHEX(%s), UNHEX(%s), UNHEX(%s))', (link_uuid, role_uuid, user_uuid))

    role = Entity(role_uuid)
    user = Entity(user_uuid)
    link = Link(link_uuid)

    for permission in kama.acl.DEFAULT_ACLS['all']:
        Permission._create(role, role, permission)
    for permission in kama.acl.DEFAULT_ACLS['all']:
        Permission._create(role, user, permission)
    
    return Entity(role_uuid), Entity(user_uuid)


class RequestContext(object):
    def __init__(self, user):
        self.user = user


class InternalContext(object):
    '''
    Instances of InternalContext bypass all permissions checks. These are used
    mostly for reading the permissions off an entity in order to determine if
    you should have permission to read it at all.
    '''
    pass


def require_permission(name):
    def decorator(fn):
        @functools.wraps(fn)
        def wrapper(self, *args, **kwargs):
            context = kwargs.get('context', None)
            if context is None:
                raise PermissionDeniedException('No context, no permission.')
            
            if isinstance(context, InternalContext):
                # Short circuit permissions test for calls that are used to test permissions
                return fn(self, *args, **kwargs)
            else:
                if context.user.has_permission(self, name):
                    return fn(self, *args, **kwargs)
                else:
                    raise PermissionDeniedException('%s does not have permission to %s %s' % (context.user, name, self))
        return wrapper
    return decorator


class Entity(object):
    def __init__(self, uuid, kind=None, name=None):
        self.uuid = uuid

        # If we don't already know the kind and name, query it.
        if kind is None or name is None:
            self.get()
        else:
            self.kind = kind
            self.name = name

    def __str__(self):
        return '%s/%s' % (self.kind, self.name)

    def __repr__(self):
        return '%s(%r, %r, %r)' % (self.__class__.__name__, self.uuid, self.kind, self.name)

    def __eq__(self, other):
        return self.uuid == other.uuid

    def __ne__(self, other):
        return self.uuid != other.uuid

    def __cmp__(self, other):
        return int(self.uuid, 16) - int(other.uuid, 16)

    @classmethod
    def create(cls, kind, name, owner_role):
        with get_database_context() as cursor:
            entity_uuid = generate_uuid()
            cursor.execute('INSERT INTO entities(entity_uuid, kind, name) VALUES(UNHEX(%s), %s, %s)', (entity_uuid, kind, name))
            self = cls(entity_uuid, kind, name)
            for permission in kama.acl.DEFAULT_ACLS['all']:
                Permission._create(owner_role, self, permission)
            return self

    @classmethod
    def get_by_name(self, kind, name):
        with get_database_context() as cursor:
            cursor.execute('SELECT HEX(entity_uuid), kind, name FROM entities WHERE name=%s AND kind=%s', (name, kind))
            row = cursor.fetchone()
            if not row:
                return None
            else:
                return Entity(*row)

    @classmethod
    def get_by_kind(self, kind):
        with get_database_context() as cursor:
            cursor.execute('SELECT HEX(entity_uuid), kind, name FROM entities WHERE kind=%s', (kind,))
            return [Entity(*x) for x in cursor]

    @classmethod
    def get_all(cls):
        with get_database_context() as cursor:
            cursor.execute('SELECT HEX(entity_uuid), kind, name FROM entities')
            return [cls(*x) for x in cursor]

    def get(self):
        with get_database_context() as cursor:
            cursor.execute('SELECT kind, name FROM entities WHERE entity_uuid=UNHEX(%s)', (self.uuid,))
            row = cursor.fetchone()
            if row is None:
                raise EntityNotFoundException(self.uuid)
            self.kind, self.name = row

    @require_permission('entity.delete')
    def delete(self, context=None):
        with get_database_context() as cursor:
            cursor.execute('DELETE FROM entities WHERE entity_uuid=UNHEX(%s)', (self.uuid,))
            cursor.execute('DELETE FROM attributes WHERE entity_uuid=UNHEX(%s)', (self.uuid,))
            cursor.execute('DELETE FROM links WHERE to_uuid=UNHEX(%s) OR from_uuid=UNHEX(%s)', (self.uuid, self.uuid))
            cursor.execute('DELETE FROM permissions WHERE entity_uuid=UNHEX(%s) OR role_uuid=UNHEX(%s)', (self.uuid, self.uuid))
        if isinstance(context, RequestContext):
            audit_log.info('%s entity.delete %s', context.user, self)

    @require_permission('entity.set_name')
    def set_name(self, name, context=None):
        with get_database_context() as cursor:
            cursor.execute('UPDATE entities SET name=%s WHERE entity_uuid=UNHEX(%s)', (name, self.uuid))
            self.name = name
        if isinstance(context, RequestContext):
            audit_log.info('%s entity.set_name %s', context.user, self)

    @require_permission('entity.add_attribute')
    def add_attribute(self, key, value, context=None):
        attribute = Attribute._create(self.uuid, key, value)
        if isinstance(context, RequestContext):
            audit_log.info('%s entity.add_attribute %s', context.user, attribute)
        return attribute

    @require_permission('entity.delete_attribute')
    def delete_attributes(self, key, context=None):
        with get_database_context() as cursor:
            cursor.execute('DELETE FROM attributes WHERE entity_uuid=UNHEX(%s) AND `key`=%s', (self.uuid, key))
            if cursor.rowcount == 0:
                raise AttributeNotFoundException('No attributes with key=%s on %s' % (key, self))
        if isinstance(context, RequestContext):
            audit_log.info('%s entity.delete_attribute key=%s', context.user, key)

    @require_permission('entity.read_attribute')
    def attributes(self, key=None, context=None):
        with get_database_context() as cursor:
            if key is not None:
                cursor.execute('SELECT HEX(attribute_uuid), `key`, `value` FROM attributes WHERE entity_uuid=UNHEX(%s) AND `key`=%s', (self.uuid, key))
            else:
                cursor.execute('SELECT HEX(attribute_uuid), `key`, `value` FROM attributes WHERE entity_uuid=UNHEX(%s)', (self.uuid,))
            if isinstance(context, RequestContext):
                audit_log.info('%s entity.read_attribute key=%s', context.user, key)
            return [Attribute(*x) for x in cursor]

    @require_permission('entity.add_link')
    def add_link(self, to_entity, context=None):
        link = Link._create(self, to_entity)
        if isinstance(context, RequestContext):
            audit_log.info('%s entity.add_link %s', context.user, link)
        return link

    @require_permission('entity.delete_link')
    def delete_link(self, to_entity, context=None):
        with get_database_context() as cursor:
            cursor.execute('DELETE FROM links WHERE from_uuid=UNHEX(%s) AND to_uuid=UNHEX(%s)', (self.uuid, to_entity.uuid))
            if cursor.rowcount == 0:
                raise LinkNotFoundException('No link from %s to %s' % (self, to_entity))
        if isinstance(context, RequestContext):
            audit_log.info('%s entity.delete_link %s -> %s', context.user, self, to_entity)

    @require_permission('entity.read_link')
    def links_from(self, context=None):
        links = []
        with get_database_context() as cursor:
            cursor.execute('SELECT HEX(links.link_uuid), HEX(links.to_uuid), entities.kind, entities.name FROM links, entities WHERE from_uuid=UNHEX(%s) AND entities.entity_uuid=links.to_uuid', (self.uuid,))
            for link_uuid, to_uuid, to_entity_kind, to_entity_name in cursor:
                to_entity = Entity(to_uuid, to_entity_kind, to_entity_name)
                link = Link(link_uuid, self, to_entity)
                links.append(link)
        if isinstance(context, RequestContext):
            audit_log.info('%s entity.read_link %s', context.user, self)
        return links

    @require_permission('entity.read_link')
    def links_to(self, context=None):
        links = []
        with get_database_context() as cursor:
            cursor.execute('SELECT HEX(link_uuid), HEX(from_uuid), entities.kind, entities.name FROM links, entities WHERE to_uuid=UNHEX(%s) AND entities.entity_uuid=links.from_uuid', (self.uuid,))
            for link_uuid, from_uuid, from_entity_kind, from_entity_name in cursor:
                from_entity = Entity(from_uuid, from_entity_kind, from_entity_name)
                link = Link(link_uuid, from_entity, self)
                links.append(link)
        if isinstance(context, RequestContext):
            audit_log.info('%s entity.read_link %s', context.user, self)
        return links

    @require_permission('entity.read_link')
    def links(self, context=None):
        return self.links_from(context=context) + self.links_to(context=context)

    @require_permission('entity.add_permission')
    def add_permission(self, role, name, context=None):
        permission = Permission._create(role, self, name)
        if isinstance(context, RequestContext):
            audit_log.info('%s entity.add_permission %s', context.user, permission)
        return permission

    @require_permission('entity.delete_permission')
    def delete_permission(self, role, name, context=None):
        with get_database_context() as cursor:
            cursor.execute('DELETE FROM permissions WHERE entity_uuid=UNHEX(%s) AND role_uuid=UNHEX(%s) AND name=%s', (self.uuid, role.uuid, name))
        if isinstance(context, RequestContext):
            audit_log.info('%s entity.delete_permission %s %s %s', context.user, self, role, name)

    @require_permission('entity.read_permission')
    def permissions(self, context=None):
        permissions = []
        with get_database_context() as cursor:
            cursor.execute('SELECT HEX(permission_uuid), HEX(role_uuid), entities.kind, entities.name, permissions.name FROM permissions, entities WHERE permissions.entity_uuid=UNHEX(%s) AND entities.entity_uuid=permissions.role_uuid', (self.uuid,))
            for permission_uuid, role_uuid, role_kind, role_name, name in cursor:
                role = Entity(role_uuid, role_kind, role_name)
                permission = Permission(permission_uuid, role, self, name)
                permissions.append(permission)
        if isinstance(context, RequestContext):
            audit_log.info('%s entity.read_permission %s', context.user, self)
        return permissions

    def has_permission(self, entity, name):
        with get_database_context() as cursor:
            cursor.execute('SELECT HEX(permissions.permission_uuid) FROM permissions, entities AS role, entities AS users, links WHERE permissions.entity_uuid=UNHEX(%s) AND permissions.name=%s AND role.entity_uuid=permissions.role_uuid AND links.from_uuid=role.entity_uuid AND links.to_uuid=users.entity_uuid AND users.entity_uuid=UNHEX(%s)', (entity.uuid, name, self.uuid))
            if cursor.rowcount > 0:
                return True
            else:
                return False

class Attribute(object):
    def __init__(self, uuid, entity_uuid=None, key=None, value=None):
        self.uuid = uuid

        if entity_uuid is None or key is None or value is None:
            self.get()
        else:
            self.entity_uuid = entity_uuid
            self.key = key
            self.value = value

    def __str__(self):
        return '%s %s=%s' % (self.entity(), self.key, self.value)

    def __repr__(self):
        return 'Attribute(%r, %r, %r, %r)' % (self.uuid, self.entity_uuid, self.key, self.value)

    def __eq__(self, other):
        return self.uuid == other.uuid

    def __ne__(self, other):
        return self.uuid != other.uuid

    def __cmp__(self, other):
        return int(self.uuid, 16) - int(other.uuid, 16)

    @classmethod
    def _create(cls, entity_uuid, key, value):
        with get_database_context() as cursor:
            attribute_uuid = generate_uuid()
            cursor.execute('INSERT INTO attributes(attribute_uuid, entity_uuid, `key`, `value`) VALUES(UNHEX(%s), UNHEX(%s), %s, %s)',
                    (attribute_uuid, entity_uuid, key, value))
            return cls(attribute_uuid, entity_uuid, key, value)

    def get(self):
        with get_database_context() as cursor:
            cursor.execute('SELECT HEX(entity_uuid), `key`, `value` FROM attributes WHERE attribute_uuid=UNHEX(%s)', (self.uuid,))
            row = cursor.fetchone()
            if row is None:
                raise AttributeNotFoundException(self.uuid)
            self.entity_uuid, self.key, self.value = row

    def entity(self):
        return Entity(self.entity_uuid)


class Link(object):
    def __init__(self, uuid, from_entity=None, to_entity=None):
        self.uuid = uuid

        if from_entity is None or to_entity is None:
            self.get()
        else:
            self.from_entity = from_entity
            self.to_entity = to_entity

    def __str__(self):
        return 'Link %s -> %s' % (self.from_entity, self.to_entity)

    def __repr__(self):
        return 'Link(%r, %r, %r)' % (self.uuid, self.from_entity, self.to_entity)

    def __eq__(self, other):
        return self.uuid == other.uuid

    def __ne__(self, other):
        return self.uuid != other.uuid

    def __cmp__(self, other):
        return int(self.uuid, 16) - int(other.uuid, 16)

    @classmethod
    def _create(cls, from_entity, to_entity):
        with get_database_context() as cursor:
            link_uuid = generate_uuid()
            cursor.execute('INSERT INTO links (link_uuid, from_uuid, to_uuid) VALUES(UNHEX(%s), UNHEX(%s), UNHEX(%s))',
                    (link_uuid, from_entity.uuid, to_entity.uuid))
            return cls(link_uuid, from_entity, to_entity)

    def get(self):
        with get_database_context() as cursor:
            cursor.execute('SELECT HEX(from_uuid), HEX(to_uuid) FROM links WHERE link_uuid=UNHEX(%s)', (self.uuid,))
            row = cursor.fetchone()
            if row is None:
                raise LinkNotFoundException(self.uuid)
            from_uuid, to_uuid = row
            self.from_entity = Entity(from_uuid)
            self.to_uuid = Entity(to_uuid)


class Permission(object):
    def __init__(self, uuid, role=None, entity=None, name=None):
        self.uuid = uuid

        if role is None or entity is None or name is None:
            self.get()
        else:
            self.role = role
            self.entity = entity
            self.name = name

    def __str__(self):
        return 'Permission %s has %s on %s' % (self.role, self.name, self.entity)

    def __repr__(self):
        return 'Permission(%r, %r, %r, %r)' % (self.uuid, self.role, self.entity, self.name)

    def __eq__(self, other):
        return self.uuid == other.uuid

    def __ne__(self, other):
        return self.uuid != other.uuid

    def __cmp__(self, other):
        return int(self.uuid, 16) - int(other.uuid, 16)

    @classmethod
    def _create(cls, role, entity, name):
        with get_database_context() as cursor:
            permission_uuid = generate_uuid()
            cursor.execute('INSERT INTO permissions (permission_uuid, role_uuid, entity_uuid, name) VALUES(UNHEX(%s), UNHEX(%s), UNHEX(%s), %s)',
                    (permission_uuid, role.uuid, entity.uuid, name))
            return cls(permission_uuid, role, entity, name)
    
    def get(self):
        with get_database_context() as cursor:
            cursor.execute('SELECT HEX(role_uuid), HEX(entity_uuid), name FROM permissions WHERE permission_uuid=UNHEX(%s)', (self.uuid,))
            row = cursor.fetchone()
            if row is None:
                raise PermissionNotFoundException(self.uuid)
            role_uuid, entity_uuid, self.name = row
            self.role = Entity(role_uuid)
            self.entity = Entity(entity_uuid)

    def users(self, context=None):
        return [x.from_entity for x in self.role.links_to(context=context)]
